#
# NeoScript - Server Side Programming based on Safe Tcl
#
# Copyright (C) NeoSoft, All Rights Reserved.
#
# $Id: init.tcl,v 1.58 1996/06/28 20:30:57 kunkee Exp $
#

#load libneo7.5.so

# load in shared base functions
source $server(SERVER_ROOT)/conf/common.tcl
#
# If we're debugging, we load devel.tcl every time we need to
# create a safe interpreter.
#
# If we're not debugging, we load devel.tcl at startup.
#
# devel_setup is executed every time a safe interpreter is
# created, in either case.
#

set debugging 1

if !$debugging {
    uplevel #0 source $server(SERVER_ROOT)/conf/devel.tcl
}

proc SAFE_dump_environment {safeInterp} {
    global webenv

    html "<pre>"
    foreach var [lsort [array names webenv]] {
        html "$var = $webenv($var)\n"
    }
    html "</pre>"
}

proc SAFE_parray {safeInterp arrayName} {
    html "<pre><b>$arrayName</b>\n"
    foreach element [lsort [$safeInterp eval array names $arrayName]] {
	html "$element = [$safeInterp eval set ${arrayName}($element)]\n"
    }
    html "</pre>"
}

proc SAFE_backlink {safeINterp {linktext ""} {nolinktext ""}} {
    global webenv

    if ![info exists webenv(HTTP_REFERER)] {
	return $nolinktext
    }

    if {$linktext == ""} {
	set linktext "Back"
    }
    return "<a href=$webenv(HTTP_REFERER)>$linktext</a>"
}

#
# PROPERTY LIST STORAGE AND RETRIEVAL
#

proc SAFE_dbfetch {safeInterp database id arrayName} {
    global parallelDir errorCode errorInfo

    setup_data_access

    db_name_check $database db dbFileName

    if {[catch {set db [db open $dbFileName hash rwl 0664]} result] == 1} {
	if {[lrange $errorCode 0 1] == "POSIX ENOENT"} {
	    return 0
	} else {
	    error "$dbFileName: $errorCode" $errorInfo
	}
    }
    db get $db $id list
    db close $db

    set haveData [expr ![lempty $list]]
    $safeInterp eval array set $arrayName [list $list]

    return $haveData
}

proc SAFE_dbstore {safeInterp database id arrayName} {
    global parallelDir errorCode errorInfo

    setup_data_access

    db_name_check $database db dbFileName

    set list [$safeInterp eval array get $arrayName]

    if {[catch {set db [db open $dbFileName hash rwL 0664]}] == 1} {
	if {[lrange $errorCode 0 1] != "POSIX ENOENT"} {
	    error "$dbFileName: $errorCode" $errorInfo
	} else {
	    create_user_dir db
	    set db [db open $dbFileName hash ctL 0664]
	}
    }
    db put $db $id $list
    db close $db
    return
}

proc SAFE_load_cookies {safeInterp {cookieVar "cookies"}} {
    global webenv

    if ![info exists webenv(HTTP_COOKIE)] return

    foreach pair [split $webenv(HTTP_COOKIE) ";"] {
	set pair [split [string trim $pair] "="]
	set key [lindex $pair 0]
	set value [lindex $pair 1]
	$safeInterp eval set ${cookieVar}($key) [list $value]
    }
}

#
# Crack the response (we're running as a pseudo-CGI in this case)
# into a global array of key-value pairs in the safe interpreter.
#
# Note info exists may return true when there is no data.
# We could check the action here and see if it's a POST
# and then we'd know which variable to look at.  The whole
# thing is somewhat kludgey.
#
# Now that we are supporting MIME types to handle file uploads, we
# need to find out if the response is mime, crack it out, and then
# handle it correctly.  Seek.  Locate.  Exterminate.
proc SAFE_load_response {safeInterp {responseVar "response"} {multiple ""}} {
    global webenv

    set responseString ""

    if {[info exists webenv(QUERY_STRING)]} {
	set responseString $webenv(QUERY_STRING)
    }

    if {[info exists webenv(NEO_POST_DATA)]} {
	set responseString $webenv(NEO_POST_DATA)
    }

    foreach key $multiple {
	set response(__$key) {}
    }

    foreach pair [split $responseString "&"] {
	set pair [split $pair "="]
	set key [www_unescape_string [lindex $pair 0]]
	set value [www_unescape_string [lindex $pair 1]]
	if [info exists response(__$key)] {
	    lappend response($key) $value
	} else {
	    if [info exists response($key)] {
		set response(__$key) {}
		set response($key) [list $response($key) $value]
	    } else {
		set response($key) $value
	    }
	}
    }
    $safeInterp eval array set $responseVar [list [array get response]]
}

#
# Increment a page counter based on the DOCUMENT_URI and return the result.
# Start it at zero if it doesn't exist (i.e. it returns 1 on the first
# try.)
#
proc SAFE_incr_page_counter {safeInterp} {
    global webenv

    if {[catch {dbfetch pagecounters $webenv(DOCUMENT_URI) accessInfo}] == 1} {
	set accessInfo(counter) 0
    }
    if ![info exists accessInfo(counter)] {
	set accessInfo(counter) 0
    }
    set counter [incr accessInfo(counter)]
    dbstore pagecounters $webenv(DOCUMENT_URI) accessInfo
    return $counter
}

#
# Pick an element in a list at random and return it
#
proc random_pick_html {list} {
    return [lindex $list [expr [clock seconds] % [llength $list]]]
}

#
# Log a message to the log directory we keep for the users.
#
#
proc SAFE_log_message {safeInterp logfile string} {
    global parallelDir errorCode errorInfo webenv

    setup_data_access

    db_name_check $logfile log logFileName

    if {[catch {set fp [open $logFileName "CREAT WRONLY APPEND" 0660]} result] == 1} {
	if {[lrange $errorCode 0 1] != "POSIX ENOENT"} {
	    error "$logFileName: $errorCode"
	}
	create_user_dir log
	set fp [open $logFileName "CREAT WRONLY APPEND" 0660]
    }
    puts $fp [list [clock seconds] $webenv(REMOTE_HOST) $string]
    close $fp
    return
}

#
# Access a data file with all the file stuff present in safe interpreters.
#
proc SAFE_access_data_file {safeInterp datafile} {
    global parallelDir errorCode
    setup_data_access

    db_name_check $datafile data dataFileName

    if {[catch {set fp [open $dataFileName "CREAT RDWR" 0660]}] == 1} {
	if {[lrange $errorCode 0 1] != "POSIX ENOENT"} {
	    error "$logFileName: $errorCode"
	}
	create_user_dir data
	set fp [open $dataFileName "CREAT RDWR" 0660]
    }

    interp transfer "" $fp $safeInterp
    return $fp
}

proc SAFE_delete_data_file {safeInterp datafile} {
    global parallelDir errorCode
    setup_data_access

    db_name_check $datafile data dataFileName

    if {[catch {unlink $dataFileName} result] == 1} {
	if {[lrange $errorCode 0 1] != "POSIX ENOENT"} {
	    error "$result"
	}
    }
    return
}

#
# Send mail
#
proc SAFE_open_outbound_mail {safeInterp subject {to ""}} {
    global webpageOwnerName webenv

    setup_data_access

    set from "$webpageOwnerName@$webenv(SERVER_NAME)"

    if {$to == ""} {
	set to $from
    }

    set fp [open "|/usr/sbin/sendmail -t" w]
    puts $fp "From: $from"
    puts $fp "To: $to"
    puts $fp "Subject: $subject"
    puts $fp "X-Mailer: $webenv(SERVER_SOFTWARE)"
    puts $fp "X-Webserver-Host: $webenv(SERVER_NAME)"
    puts $fp "X-Sender-URI: $webenv(DOCUMENT_URI)"
    puts $fp ""

    interp transfer "" $fp $safeInterp
    return $fp
}

#
# Post news
#

#Newsgroups: comp.lang.tcl
#Subject: Foo
#Summary:
#Followup-To:
#Distribution:
#Organization: NeoSoft, Inc. +1 713 968 5800
#Keywords:
#Cc:

proc SAFE_open_post_news {safeInterp args} {
    global webpageOwnerName webenv

    import_keyvalue_pairs data $args

    if ![info exists data(newsgroups)] {
	error "-newsgroups not specified"
    }

    if ![info exists data(subject)] {
	error "-subject not specified"
    }

    if ![info exists data(distribution)] {
	set data(distribution) ""
    }

    setup_data_access

    set from "$webpageOwnerName@[remote_host $webenv(SERVER_NAME)]"

    set fp [socket news 119]
    gets $fp line
    if {[string index $line 0] != 2} {
	error "news server not ready: $line"
    }
    puts $fp "post"
    flush $fp
    gets $fp line
    if {[string index $line 0] != "3"} {
	error "news server post command failed: $line"
    }

    puts $fp "From: $from"
    puts $fp "Newsgroups: $data(newsgroups)"
    puts $fp "Subject: $data(subject)"
    puts $fp "Distribution: $data(distribution)"
    puts $fp {X-News-Posting-Program: neoscript-$Date: 1996/06/28 20:30:57 $-$Version$}
    puts $fp ""

    interp transfer "" $fp $safeInterp
    return $fp
}

proc SAFE_directory_listing {safeInterp} {

    set duFP [open "|du -s" r]
    gets $duFP line
    close $duFP

    set total [lindex $line 0]
    html "Total space used from this level down is $total Kbytes."

    html "<table border>"
    html "<tr><th>name</th><th>type</th><th>size</th><th>last mod</th></tr>"
    html "<tr><td><a href=..>..</a></td><td>directory</td><td>.</td><td>.</td>"
    foreach file [lsort [glob -nocomplain *]] {
        if {[catch {file stat $file x} result] == 1} {
	    html "<tr><td>$file</td><td>$result</td></tr>"
	} else {
	    set date [clock format $x(mtime)]
	    html [format "<tr><td><a href=%s>%s</a></td><td>%s</td><td align=right>%s</td><td>%s</td></tr>\n" $file $file $x(type) $x(size) $date]
	}
    }
    html "</table>"
}

#
# guess at the number of hits we're getting
# per hour by calculating the rate of growth
# of the log file.
#
# assumes logs are in SERVER_ROOT/logs/access_log
#
proc estimate_hits_per_hour {} {
    global server

    set hitsInInterval 1000
    set bytesPerLogEntry 92

    set fp [open "$server(SERVER_ROOT)/logs/access_log"]
    set size [fstat $fp size]
    set offset [expr $size - $hitsInInterval * $bytesPerLogEntry]
    if {$offset > 0} {
	seek $fp $offset
	gets $fp
    } else {
	set hitsInInterval [expr $size / $bytesPerLogEntry]
    }

    set result [gets $fp line]
    close $fp
    if {$result < 0} {return 0}

    set ET [expr [clock seconds] - [lindex $line 0]]
    return [expr ($hitsInInterval * 3600) / $ET]
}

#
# Return the hostname of the machine that is fetching the current
# page.  If we don't have the name because they're running the
# server with minimal DNS, perform a DNS lookup for the hostname
# and return that.
#
# If we can't figure out the hostname, we return the ip address.
#
proc remote_hostname {{ip 0}} {
    global webenv

    if {$ip != 0} {
       if {[catch {host_info official_name $ip} result] == 1} {
	   return $ip
       }
    } else {
       if {$webenv(REMOTE_HOST) != $webenv(REMOTE_ADDR)} {
	   return $webenv(REMOTE_HOST)
       }
       if {[catch {host_info official_name $webenv(REMOTE_ADDR)} result] == 1} {
	   return $webenv(REMOTE_ADDR)
       }
    }
    return $result
}


#
# destroy the safe interpreter if it already exists, then
# create a new one and install our services (exported
# procs) into it
#
proc setup_safe_interpreter {} {
    global debugging parallelDir errorInfo NeoWebDirConf

    if [info exists parallelDir] {
        unset parallelDir
        set errorInfo ""
    }

    if {[info exists NeoWebDirConf(Supervisor)] && $NeoWebDirConf(Supervisor)} {
	set safeInterp [interp create]
	load {} Neo $safeInterp
    } else {
	set safeInterp [interp create -safe]
    }

    load {} Tclx $safeInterp

    $safeInterp alias dump_environment SAFE_dump_environment $safeInterp
    $safeInterp alias parray SAFE_parray $safeInterp
    $safeInterp alias backlink SAFE_backlink $safeInterp

    $safeInterp alias dbfetch SAFE_dbfetch $safeInterp
    $safeInterp alias dbstore SAFE_dbstore $safeInterp
    $safeInterp alias dbkeys dbkeys
    $safeInterp alias dbexists dbexists
    $safeInterp alias dbdelkey dbdelkey
    $safeInterp alias dbdump dbdump
    $safeInterp alias dbfiles dbfiles

    $safeInterp alias log_message SAFE_log_message $safeInterp

    $safeInterp alias access_data_file SAFE_access_data_file $safeInterp
    $safeInterp alias delete_data_file SAFE_delete_data_file $safeInterp

    $safeInterp alias list_data_files list_data_files

    $safeInterp alias open_outbound_mail SAFE_open_outbound_mail $safeInterp
    $safeInterp alias open_post_news SAFE_open_post_news $safeInterp

    $safeInterp alias load_cookies SAFE_load_cookies $safeInterp
    $safeInterp alias load_response SAFE_load_response $safeInterp

    $safeInterp alias random_pick_html random_pick_html
    $safeInterp alias abort_page abort_page

    $safeInterp alias html html

    $safeInterp alias include_file SAFE_include_file $safeInterp
    $safeInterp alias include_virtual SAFE_include_virtual $safeInterp
    $safeInterp alias load_file SAFE_load_file $safeInterp
    $safeInterp alias load_virtual SAFE_load_virtual $safeInterp

    $safeInterp alias unquote_string www_unescape_string
    $safeInterp alias quote_string www_escape_string

    $safeInterp alias directory_listing SAFE_directory_listing $safeInterp
    $safeInterp alias du SAFE_du

    $safeInterp alias estimate_hits_per_hour estimate_hits_per_hour
    $safeInterp alias remote_hostname remote_hostname

    extend_slave $safeInterp

    if $debugging {
	global server
	uplevel #0 source $server(SERVER_ROOT)/conf/devel.tcl
    }

    if [catch {devel_setup $safeInterp} errorMsg] {
	html {devel_setup error}
	global errorInfo
	html $errorInfo code
    }

    setup_slave_interp_unknown $safeInterp

    return $safeInterp
}

#
# The code called from the interface from apache
#
# safeInterp is the interpreter we created for
# our mod_neoinclude C module when it asked us to
# (setup_safe_interpreter, above)
#
# tag and tag_val are the key-value pairs in the server
# side include
#
proc handle_neoscript_request {safeInterp tag tag_val} {
    global debugging errorInfo

    switch $tag {
	"return" {
	    handle_server_side_return $safeInterp $tag_val
	}
	"code" {
	    handle_server_side_eval $safeInterp $tag_val
	}
	"eval" {
	    handle_server_side_eval $safeInterp $tag_val
	}
	"var" {
	    handle_server_side_variable $safeInterp $tag_val
	}
	"expr" {
	    handle_server_side_expr $safeInterp $tag_val
	}
	default {
	    html "\[Error -- unrecognized tag in neoscript directive\]"
	}
    }

    return
}

#
# The code called from handle_neoscript_request to do an eval=
#
proc handle_server_side_eval {safeInterp ssiCode} {
    global debugging errorInfo

    if {[catch {$safeInterp eval $ssiCode} result] == 1} {
	html "\{NEOSCRIPT ERROR: $result\}"
	if {$debugging} {
	    html "\n<pre>$errorInfo</pre>"
	}
    }
    return
}

#
# The code called from handle_neoscript_request to do a return=
#
proc handle_server_side_return {safeInterp ssiCode} {
    global debugging errorInfo

    if {[catch {$safeInterp eval $ssiCode} result] == 1} {
	html "\{NEOSCRIPT ERROR: $result\}"
	if {$debugging} {
	    html "\n<pre>$errorInfo</pre>"
	}
    } else {
	if {$result != ""} {
	    html $result
	}
    }
    return
}

#
# The code called from handle_neoscript_request to do a var=
#
proc handle_server_side_variable {safeInterp ssiVar} {
    global debugging

    if {[catch {$safeInterp eval set $ssiVar} result] == 1} {
	html "\{NEOSCRIPT ERROR: $result\}"
	if {$debugging} {
	    html "\n<pre>$errorInfo</pre>"
	}
    } else {
	if {$result != ""} {
	    html $result
	}
    }
    return
}

#
# The code called from handle_neoscript_request to do a var=
#
proc handle_server_side_expr {safeInterp ssiExpr} {
    global debugging

    if {[catch {$safeInterp eval expr $ssiExpr} result] == 1} {
	html "\{NEOSCRIPT ERROR: $result\}"
	if {$debugging} {
	    html "\n<pre>$errorInfo</pre>"
	}
    } else {
	if {$result != ""} {
	    html $result
	}
    }
    return
}

proc setup_slave_interp_unknown {safeInterp} {
    # Experimental implementation of unknown
    global webenv SafePaths tcl_library NeoWebDirConf UnknownTcl
    $safeInterp eval $UnknownTcl
    set local_lib_path [www_simplify_pathname [file dirname $webenv(SCRIPT_FILENAME)]]
    set SafePaths "$tcl_library $webenv(SERVER_ROOT)/neoscript-tcl $local_lib_path"
    $safeInterp eval set auto_path [list $SafePaths]
    if {![info exists NeoWebDirConf(Supervisor)] || ! $NeoWebDirConf(Supervisor)} {
	$safeInterp alias source SAFE_source $safeInterp
	$safeInterp alias open SAFE_open $safeInterp
	$safeInterp alias file file
    }
}

proc read_file {filename} {
	set fp [open $filename]
	set data [read $fp]
	close $fp
	return $data
}

set UnknownTcl [read_file $server(SERVER_ROOT)/conf/unknown.tcl]
eval $UnknownTcl
lappend auto_path $server(SERVER_ROOT)/neoscript-tcl

proc SAFE_filename {filename} {
    global webenv SafePaths
    if {[cindex $filename 0] != "/"} {
	set filename [file join [file dirname $webenv(SCRIPT_FILENAME)] $filename]
    }
    set filename [www_simplify_pathname $filename]
    if [lsearch -exact $SafePaths [file dirname $filename]]<0 {
	return
    }
    return $filename
}

proc SAFE_open {safeInterp filename args} {
    set nfilename [SAFE_filename $filename]
    if [lempty $nfilename] {
	error "$filename: Invalid pathname requested"
    }
    set channel [open $nfilename]
    interp transfer {} $channel $safeInterp
    return $channel
}

proc SAFE_source {safeInterp filename} {
    set nfilename [SAFE_filename $filename]
    if [lempty $nfilename] {
	error "$filename: Invalid pathname requested"
    }
    $safeInterp eval [read_file $nfilename]
}
